"""
Creates a randomised maze matrix that can be converted into an image.
It is assumed that the global maze matrix is stored at ./g.py:maze.

The maze is compatible with mazesolver (https://github.com/exciteabletom/mazesolver)

There are two rules for the output matrix that make it compatible:

 - Walls around the entire maze
 - One entrance on the top row and one exit on the bottom row

The maze is represented in the matrix as follows:
 - Walls:   "#"
 - Paths:   "."
 - Start:   "s"
 - End:     "e"

Example output (5x5):

   [["#", "#", "#", "s", "#"],
    ["#", "#", ".", ".", "#"],
    ["#", ".", "#", ".", "#"],
    ["#", ".", ".", ".", "#"],
    ["#", ".", "#", ".", "#"],
    ["#", "e", "#", "#", "#"]]
"""
## generate.py - Tommy Dougiamas

# Standard libraries
import os
import random

# Third party
import progress.bar  # Progress bars

# Relative
from . import mazeutils as mu
from . import g


def check_seed():
	"""
	Creates a random seed if one is not defined already
	"""
	if not g.seed:  # If no user-defined seed
		# Create random seed
		random_chars = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s",
                            "t", "u", "v", "w" "x", "y", "z", "1", "2", "3", "4", "5", "6", "7", "8", "9", "0"]
		for _ in range(15):
			rand_bool = random.random() < 0.5
			rand_char = random_chars[random.randint(0, len(random_chars) - 1)]

			if rand_bool:
				rand_char = rand_char.upper()

			g.seed = g.seed + rand_char

	print(f"Using seed '{g.seed}'\n")
	random.seed(g.seed)


def init_maze(width, height):
	"""
	Initialises a maze with only walls
	:param width: The width of the maze
	:param height: The height of the maze
	"""
	g.maze = []

	progress_bar = progress.bar.PixelBar(g.change_string_length("Initialising empty maze", 30), max=height)

	for _ in range(height):
		progress_bar.next()
		g.maze.append([])
		for _ in range(width):
			g.maze[-1].append("#")

	progress_bar.finish()


def branch(coords: tuple, direction: str, no_exit: bool = False, noise_offset: float = 0.0):
	"""
	Branches out to the side of a target cell, either left, right or down, used to add tree like structure

	:param coords: (x,y) indicating a cell position
	:param direction: 'left', 'right' or 'down'
	:param no_exit: Bool indicating whether to not stop randomly
	:param noise_offset: float that affects some of the random chances
	:return: The cell that was last visited
	:rtype: tuple
	"""

	while True:
		rand_float = random.random() + noise_offset
		if rand_float < 0.05 and not no_exit:
			return coords

		neighbour_directions = mu.get_cell_neighbour_direction_names(coords, empty_cell="#")

		if direction in neighbour_directions:
			final_direction = direction
			if 0.05 < rand_float < 0.45 + noise_offset:
				final_direction = "down"

			try:
				next_coords = mu.get_cell_neighbours(coords, "#", final_direction)[0]
			except IndexError:
				return coords

			if not mu.is_edge(next_coords):
				if next_coords[0] == len(g.maze) - 1:
					breakpoint()
				mu.set_cell_value(next_coords, ".")
				coords = next_coords
			else:
				return coords
		else:
			return coords


def init_solution_path():
	"""
	Creates a randomized solution path through the maze.
	"""

	# Find the beginning of the maze
	start_pos = random.randint(1, len(g.maze[0]) - 2)
	g.maze[0][start_pos] = "s"
	start = mu.get_cell_by_value("s")

	# Set the current cell to be the cell under start
	current_cell = (start[0] + 1, start[1])
	mu.set_cell_value(current_cell, ".")

	# TODO: Implement the possibility of the path going up
	# Currently no_up will always be True meaning the path can never go upwards
	no_up = True

	if random.random() < 0.5:
		h_prefer = "right"
		not_h_prefer = "left"
	else:
		h_prefer = "left"
		not_h_prefer = "right"

	rows = len(g.maze) - 2
	last_row = 0

	progress_bar = progress.bar.PixelBar(g.change_string_length("Generating random solution", 30), max=rows)

	# Path from start
	while True:
		if current_cell[0] != last_row:
			last_row = current_cell[0]
			progress_bar.next()

		if current_cell[0] == len(g.maze) - 2:  # If on second last row of maze
			mu.set_cell_value((len(g.maze) - 1, current_cell[1]), "e")
			break

		# Possible directions we could travel to
		directions = mu.get_cell_neighbour_direction_names(current_cell, empty_cell="#")

		if no_up and "up" in directions:  # Currently will always be triggered
			directions.remove("up")

		# A random direction
		rand_direction = directions[random.randint(0, len(directions) - 1)]

		if h_prefer in directions and random.random() < 0.6:
			rand_direction = h_prefer

		elif random.random() < 0.01:
			current_cell = branch(current_cell, h_prefer)
			progress_bar.next(current_cell[0] - last_row)
			last_row = current_cell[0]
			if random.random() < 0.5:
				h_prefer, not_h_prefer = (not_h_prefer, h_prefer)
			continue

		next_cell = mu.get_cell_neighbours(current_cell, "#", rand_direction)[0]
		mu.set_cell_value(next_cell, ".")

		if mu.next_to_edge(current_cell):
			if random.random() < 0.60:
				h_prefer, not_h_prefer = (not_h_prefer, h_prefer)

		current_cell = next_cell

	progress_bar.finish()


def expand_rows(noise_offset: float):
	"""
	'expands' rows by adding random paths on, above, and below the rows
	:param noise_offset: An offset applied to some of the random float values generated
	                        A negative offset reduces noise, a positive one increases noise
	"""
	progress_bar = progress.bar.PixelBar(g.change_string_length("Adding noise", 30), max=len(g.maze))

	for row_index, row in enumerate(g.maze):
		progress_bar.next()
		if row_index % 3 == 0:
			continue

		if row_index == len(g.maze) - 1:
			continue

		for cell_index, cell in enumerate(row):
			if cell_index in (0, len(g.maze[0]) - 1):
				continue

			cell_coords = (row_index, cell_index)
			rand = random.randint(0, 13)

			if cell == "#":  # If cell is wall
				cell_neighbours = mu.get_cell_neighbours(cell_coords, empty_cell=".")

				if cell_neighbours and rand < 1:
					mu.set_cell_value(cell_coords, ".")
				elif rand in (2, 3):
					rand_direction = ""

					# Rare wildcard for more randomness TODO: Maybe too expensive to compute? 
					if random.random() < 0.005:
						rand_direction = "down"
					elif rand == 2:
						rand_direction = "left"
					elif rand == 3:
						rand_direction = "right"
					else:
						raise ValueError("DEVERROR: Random integer out of range")

					branch(cell_coords, rand_direction, random.random() < 0.001, noise_offset)

	progress_bar.finish()


def generate(width: int, height: int, noise_bias: str):
	"""
	Main function that creates the maze.
	:param width: Width of the matrix
	:param height: Height of the matrix
	:param noise_bias: Either "wall", "less", "none", or and empty string indicating no bias
	"""
	check_seed()
	init_maze(width, height)
	init_solution_path()

	if noise_bias != "none":  # If we should generate noise
		noise_offset = 0

		if noise_bias == "walls":  # Draw less paths
			print("Creating more walls")
			noise_offset = -0.09

		elif noise_bias == "paths":  # Draw more paths
			print("Creating more paths")
			noise_offset = 0.25

		expand_rows(noise_offset)
	else:
		print("Only rendering solution path")



## create_final_image.py - Tommy Dougiamas
"""
Converts a matrix into a black and white image, and saves it to a specified directory
"""

from PIL import Image  # Pillow >=6.0
from pathlib import Path  # OS agnostic filesystem paths


def create(matrix: list, output_dir: str, output_name: str):
	"""
	Void function that marks the solution path into the image with green and saves the image.

	:param matrix: A matrix generated by generate.py (see generate.__doc__).
	:param output_dir: String with User-supplied path to a directory where the image will be saved.
	:param output_name: A name for the image file
	"""
	# open the image that was inputted
	output_image = Image.new("RGB", [len(matrix[0]), len(matrix)], (255, 255, 255))

	for y in range(0, len(matrix)):
		for x in range(0, len(matrix[-1])):
			cell = matrix[y][x]

			color = (0, 0, 0)  # By default color is black

			if cell == "." or cell == "s" or cell == "e":
				color = (255, 255, 255)  # Change color to black

			output_image.putpixel((x, y), color)

	out_path = Path(f"{output_dir}/{output_name}.jpg")  # Where the image will be saved to

	print("\nSaving Image... This may take a long time for bigger mazes")

	output_image.save(out_path, subsampling=0, quality=100)  # Save the image with no compression or sub-sampling

	print(f"Maze was saved at {out_path}")  # Make sure the user knows where the image was saved



## mazeutils.py - Tommy Dougiamas
"""
Some basic utilites for manipulating and displaying the variable g.maze.
"""
import random

from . import g

maze = None


def print_maze():
	"""
	Prints out the maze matrix in a human readable format, useful for debugging.
	"""
	for i in g.maze:
		print(i)
	print("\n")


def get_cell_value(coords: tuple):
	"""
	Gets the value of the cell at the specified coordinates

	:param coords: tuple containing x and y values
	:return: value of the cell at the specifed coordinates
	"""
	try:
		return g.maze[coords[0]][coords[1]]
	# Sometimes we get an IndexError if the maze doesn't have borders
	# This solution is not perfect, so it is still best practice to use borders
	except IndexError:
		return False


def get_cells_by_value(value):
	"""
	Get cell coordinates based on the value of the cell.

	:param value: The value to search cells for
	:return: list of all coordinates that contain the specified value
	"""
	all_matching_cells = []  # the list containing all the coordinates of cells
	for row_index, row in enumerate(g.maze):
		for column_index, cell in enumerate(row):
			if cell == value:
				all_matching_cells.append((row_index, column_index))

	return all_matching_cells


def is_edge(coords):
	"""
	Check if a piece is an edge or not.

	:param coords: A tuple (x,y)
	:return: True if piece is an edge piece False otherwise
	"""
	if coords[0] == 0 or coords[0] == len(g.maze) - 1 \
		  or coords[1] == 0 or coords[1] == len(g.maze[0]) - 1:  # if edge piece
		return True

	return False


def get_cell_by_value(value):
	"""
	The same as get_cells_by_value, except raises a ValueError if there is more than one cell with that value

	:param value: The value to search cells for
	:raises ValueError: If more then one of the value is found in the maze.
	:return: the cell coordinate that contains the value
	"""
	values = get_cells_by_value(value)
	if len(values) > 1:
		raise ValueError(f"Expected only one cell to have value '{value}'. {len(values)} cells contained the value.")

	return values[0]


def set_cell_value(coords: tuple, value: any):
	"""
	Sets the value of a cell at the specified coordinates.

	:param coords: The coordinates of the cell to be changed
	:param value: The value we want the cell to be set to
	"""
	g.maze[coords[0]][coords[1]] = value


def check_cell_exists(coords):
	"""
	Checks if a cell exists within the maze.

	:param coords: A tuple (x,y), representing a cell
	:return bool: True if cell exists, False otherwise
	"""
	try:
		_ = g.maze[coords[0]][coords[1]]  # Will throw IndexError if the cell is out of the maze area
		return True  # Cell exists
	except IndexError:
		return False  # Cell doesn't exist


def get_cell_neighbours(coords: tuple, empty_cell: str = None, directions: str = None):
	"""
	Gets the values of all cells that neighbour the cell at the specified coordinates

	:param coords: Tuple containing the x and y values of the cell to check the neighbours of
	:param empty_cell: specifies an empty cell as a string
	:param directions: String containing directions to be checked for.
	:return: coordinates of all neighbours that have not been visited in
				a list of tuples. Example: [(x,y), (x,y), (x,y)]
	"""
	# different tuples that contain the coords of all positions
	# relative to our input tuple
	up = (coords[0] - 1, coords[1])
	down = (coords[0] + 1, coords[1])
	left = (coords[0], coords[1] - 1)
	right = (coords[0], coords[1] + 1)

	# list containing all directional tuples
	all_dirs = [up, down, right, left]
	if directions:
		all_dirs = []
		if "up" in directions:
			all_dirs.append(up)
		if "down" in directions:
			all_dirs.append(down)
		if "right" in directions:
			all_dirs.append(right)
		if "left" in directions:
			all_dirs.append(left)

		if not all_dirs:
			raise ValueError(f"Directions {directions} not recognised.")

	visitable_coordinates = []

	if type(empty_cell) == str:
		for dir in all_dirs:
			cell_value = get_cell_value(dir)

			if cell_value == empty_cell:
				if is_edge(dir):
					continue

				if dir[0] < 0 or dir[1] < 0:  # If negative number
					continue

				visitable_coordinates.append(dir)  # Don't remove

	return visitable_coordinates


def get_cell_neighbour_direction_names(coords, direction="all", empty_cell="."):
	"""
	Checks which directions can be moved to.

	:param coords: A tuple (x,y).
	:param direction: String containing a directions to check. If left out will check every directions.
	:param empty_cell: What value is considered empty
	:return: A list containing directions that can be moved to. E.g. ["right", "up", "left"].
	"""

	# different tuples that contain the coords of all positions
	# relative to our input tuple
	up = (coords[0] - 1, coords[1])
	down = (coords[0] + 1, coords[1])
	left = (coords[0], coords[1] - 1)
	right = (coords[0], coords[1] + 1)

	all_dirs = [(up, "up"), (down, "down"), (right, "right"), (left, "left")]
	good_dirs = []

	# The following is messy and slow TODO
	if direction == "all":
		for cell_data in all_dirs:
			if is_edge(cell_data[0]) or get_cell_value(cell_data[0]) != empty_cell:
				continue
			good_dirs.append(cell_data[1])

	else:
		if direction == "up":
			index = 0
		elif direction == "down":
			index = 1
		elif direction == "right":
			index = 2
		elif direction == "left":
			index = 3
		else:
			raise ValueError(f"Direction {direction}, not recognised.")

		if get_cell_value(all_dirs[index]) == empty_cell and not is_edge(all_dirs[index]):
			good_dirs.append(direction)

	return good_dirs


def next_to_edge(coords: tuple):
	"""
	Function for checking if a cell is next to the edge of the maze.

	:param coords: Tuple (x, y)
	:rtype: bool
	:return: True if next to edge, false otherwise
	"""
	next_to_wall: bool = False

	if coords[0] == 1 or coords[0] == len(g.maze) - 2:
		next_to_wall = True

	elif coords[1] == 1 or coords[1] == len(g.maze[-1]) - 2:
		next_to_wall = True

	return next_to_wall



## __main__.py - Tommy Dougiamas
"""
This file interprets all command line arguments.
It passes off the actual processing to other functions.
"""

# Stdlib imports
import sys
import os
from pathlib import Path  # Used to fix incompatibilities between windows and unix-based file paths ("/" vs "\\")

# Relative imports
from . import generate  # width/height --> matrix
from . import create_output_image  # matrix --> image
from . import strings  # Static strings
from . import g  # global variables


def cmd_error(message=""):  # Display error message and exit the program with exit code 1
	"""
	Displays a user-friendly error message and exits 1

	:param message: Message for error
	"""
	if message:
		print(f"ERROR: {message}\n", file=sys.stderr)

	print("See --help for more info.", file=sys.stderr)
	sys.exit(1)


def cmd_info(mode):  # Display information and exit the program with exit code 0
	"""
	Displays a user friendly information string from 'strings' and exits 0.
	"""
	if mode == "help":
		print(strings.HELP)

	elif mode == "VERSION":
		print(strings.VERSION)

	elif mode == "MAZE_RULES":
		print(strings.MAZE_RULES)

	else:
		# If the mode was not valid throw error, so then it doesn't get into prod
		raise ValueError(f"DEV_ERROR: Option '{mode}' is not valid for cmd_info. \
						If you are seeing this please contact tom@digitalnook.net")

	sys.exit(0)


def main():
	"""
	Interprets command line arguments and passes on to generate.generate()
	Then runs create_output_image.create()

	Exit code 0 if successful, code 1 if error occurs
	"""
	output_path = ""  # The path for the picture to be outputted to
	width: int = 0
	height: int = 0

	option_no_noise = False
	option_more_paths = False
	option_more_walls = False

	cmd_args = sys.argv[1:]  # List storing all command line arguments passed to the program
	if len(cmd_args) == 0:  # if no arguments were given
		cmd_error("No arguments provided.")

	# If these arguments are passed we need to display some information and exit 0
	if "--help" in cmd_args or "-h" in cmd_args:
		cmd_info("help")

	elif "-v" in cmd_args or "--version" in cmd_args:
		cmd_info("VERSION")

	elif "--maze-rules" in cmd_args:
		cmd_info("MAZE_RULES")

	skip_next_arg = False  # Boolean indicating whether the current iteration should be skipped

	# Loop handling arguments that have params like "-i" and "-o"
	# TODO: Probably a better way to handle these and still not use a library
	for index, arg in enumerate(cmd_args):
		if skip_next_arg:  # If this is true skip the current iteration
			skip_next_arg = False
			continue

		try:
			if arg in ("-o", "--output"):
				output_path = cmd_args[index + 1]  # the argument after '-o' is the output path
				skip_next_arg = True

			elif arg in ("-x", "--width"):
				width = int(cmd_args[index + 1])
				skip_next_arg = True

			elif arg in ("-y", "--height"):
				height = int(cmd_args[index + 1])
				skip_next_arg = True

			elif arg in ("-xy", "--xy"):
				height = int(cmd_args[index + 1])
				width = int(cmd_args[index + 1])
				skip_next_arg = True

			elif arg in ("--seed", "-s"):
				g.seed = cmd_args[index + 1]
				print(g.seed)
				skip_next_arg = True

			elif arg == "--no-noise":
				option_no_noise = True

			elif arg in ("--favour-walls", "--favor-walls"):
				option_more_walls = True

			elif arg in ("--favour-paths", "--favor-paths"):
				option_more_paths = True

			else:
				cmd_error(f"Option '{arg}' not recognised.")

		except IndexError:  # If no parameter is passed when an arg expects it
			cmd_error(f"Option '{arg}' requires an parameter.")

	# This block is designed to work if:
	# 1. Only a directory name is passed with or without a trailing '/' eg Pictures/ and Pictures
	# 2. An image name is passed with/without a .jpg or .jpeg extension  eg. mymaze.jpg and mymaze
	# 3. A directory name is passed with an image name  eg. Pictures/mymaze.jpg or Pictures/mymaze
	output_dir = str(Path.cwd())
	output_name = "maze"
	if output_path:
		if os.path.isdir(output_path):  # If only directory is specified
			output_dir = output_path
		else:
			path_lst = output_path.split("/")
			if path_lst[-1] == "":  # If directory was specified, but not valid
				cmd_error("Invalid directory name.")

			elif len(path_lst) == 1:  # If only image name is specified with no directory
				output_name = path_lst[0].replace(".jpg", "").replace(".jpeg", "")

			else:  # If directory and image name are specified
				output_name = path_lst[-1]
				output_dir = output_path[0:-len(output_name)]
				output_name = output_name.replace(".jpg", "").replace(".jpeg", "")

	if not width or not height:
		width = 50
		height = 50
		print("No height or width supplied, defaulting to 50x50")

	if width < 20 or height < 20:  # Generation doesn't work with super small mazes
		cmd_error("Both width and height must be at least 20.")

	noise_bias = "default"

	if option_no_noise:  # creates only a path
		noise_bias = "none"
	elif option_more_paths:
		noise_bias = "paths"
	elif option_more_walls:
		noise_bias = "walls"

	generate.generate(width, height, noise_bias)

	create_output_image.create(g.maze, output_dir, output_name)



#!/usr/bin/env python3
from setuptools import setup
import sys
from mazegenerator import strings

print(strings.VERSION)
with open("./README.md", encoding="utf-8") as readme:
	long_description = readme.read()

setup(
	name="mazegenerator",
	version=strings.VERSION,

	description="A maze generation algorithm for image-based mazes.",
	long_description=long_description,
	long_description_content_type="text/markdown",

	url="https://github.com/exciteabletom/mazegenerator",
	author="Tommy Dougiamas",
	author_email="tom@digitalnook.net",

	classifiers=[
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Programming Language :: Python :: 3"
	],

	entry_points={
		"console_scripts": ["mazegenerator = mazegenerator.__main__:main"],
	},

	keywords="maze algorithm image generate",

	packages=["mazegenerator"],

	python_requires=">=3.5",

	install_requires=["Pillow>=6.0", "progress>=1.5"]
)



## strings.py - Tommy Dougiamas
"""
This file contains all constant strings.
"""

HELP = """\
Valid Commands
---------------
-h, --help      -  Prints this help page and exits.
-v, --version   -  Prints the version of the program and exits.
--maze-rules    -  Prints the rules for maze images.

-x, --width     -  The width of the maze. (required)
-y, --height    -  The height of the maze. (required)
--xy            -  The height and width of a square maze.

--no-noise      -  Render a solution path without any noise
--favour-paths  -  Generate more paths
--favour-walls  -  Generate more walls

-s, --seed      -  Specifies a seed to be used for the random number generator
-o, --output    -  Output filepath/directory

Example Usages 
---------------
mazegenerator -x 300 -y 2000 -o path/to/dir/my_cool_maze_name
mazegenerator --xy 600 --favour-paths
mazegenerator --xy 200 -o path/to/dir/

Contact Info
---------------
Email: tom@digitalnook.net
Github: https://github.com/exciteabletom

Licensed under GPLv3: https://www.gnu.org/licenses/gpl-3.0.en.html \
"""

MAZE_RULES = """\
What are the rules for maze images?
---------------

- Walls marked with black pixels and paths marked with white pixels

- Walls around the entire maze

- One entrance on the top row and one exit on the bottom row
 
Check out some of the example mazes at https://github.com/exciteabletom/mazegenerator/tree/master/examples \
"""


VERSION = "1.2"
VERSION_LONG = f"mazegenerator{VERSION}"




import os

# Maze Matrix
maze = []

# Seed used to seed the RNG
seed = ""


def change_string_length(string, length):
	"""
	Append spaces to a string until it reaches 'length'
	"""
	diff = length - len(string)
	return string + (" " * diff)
